//And important note - enchant.js expects maps to look like this:
//[ [ 1, 2],
//  [ 3, 4]]
//  That much is normal. BUT, that means that the X dimension is the INNER
//  dimension - that is, to get the value at a position, you get a[y][x]. 
//  Confusing.
var res = 64;

var rainbow = [ [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
	        [ 0, 1, 1, 1, 1, 1, 1, 1, 1, 0],
	        [ 0, 1, 2, 2, 2, 2, 2, 2, 1, 0],
	        [ 0, 1, 2, 3, 3, 3, 3, 2, 1, 0],
	        [ 0, 1, 2, 3, 3, 3, 3, 2, 1, 0],
	        [ 0, 1, 2, 2, 2, 2, 2, 2, 1, 0],
	        [ 0, 1, 1, 1, 1, 1, 1, 1, 1, 0],
	        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]]

//Random thingy
var r = Alea("get funky", new Date());

//How many tiles is the map?
var world = {w: 10, h:8};

var gmap;

var player;
//This array holds the truth values for the rules... maybe not necessary?
var rules = new Array();
var rulecolors = new Array();
var validrules;

var colorrules;

function fillrules(a){
	//Create the fourteen rules which form the base of the game
	for(var i = 0; i< 14; i++){
		//TODO decide which directions these are
		//I think 0 is right?
		rule = [ false, false, false, false];
		if ( (i+1) & 1 ) { rule[0] = true }
		if ( (i+1) & 2 ) { rule[1] = true }
		if ( (i+1) & 4 ) { rule[2] = true }
		if ( (i+1) & 8 ) { rule[3] = true }
		a.push(rule);
	}
}

function getvalidrules(rules){
	//Takes in the list of active rules and returns an array of lists of valid
	//ones for each direction. 0 is right, and things go clockwise (down-left-up). 
	var vr = new Array();
	vr[0] = _.filter(rules, function(x){ return (x) & 1;});
	vr[1] = _.filter(rules, function(x){ return (x) & 2;});
	vr[2] = _.filter(rules, function(x){ return (x) & 4;});
	vr[3] = _.filter(rules, function(x){ return (x) & 8;});
	return vr;
}

function randomwalk(map, sx, sy, d){
	//NOTE: x and y are weird because the loadData function 
	//takes an array ordered y,x rather than x,y
	//NOTE: This modifies the map passed in
	//Params:
	//map: the array that will store the world map / dance floor
	//sx sy : starting x and y
	//d: distance to walk
	//return value: [x,y] of the goal point
	//NOTE: recursive dfs for backtracking; each call steps once and can undo
	//If we're not going anywhere we're done
	//TODO: probably need to fill the spot at sx,sy
	if(d == 0){ return [sx,sy];}
	
	//A candidate is any unvisited square that touches us that exists
	var neighbors = [];
	//Provide the x value, y value, and next rule (chosen at random)
	if(sx != map.length-1 && map[sy][sx+1] == 0){ neighbors.push([sx+1,sy,validrules[2][~~(r()*validrules[2].length)]]);}
	if(sy != 0 && map[sy-1][sx] == 0){ neighbors.push([sx,sy-1,validrules[1][~~(r()*validrules[1].length)]]);}
	if(sx != 0 && map[sy][sx-1] == 0){ neighbors.push([sx-1,sy,validrules[0][~~(r()*validrules[0].length)]]);}
	if(sy != map[0].length-1 && map[sy+1][sx] == 0){ neighbors.push([sx,sy+1,validrules[3][~~(r()*validrules[3].length)]]);}
	//randomize the order
	fys(neighbors);
	while(neighbors.length > 0){
		var n = neighbors.pop();
		map[n[1]][n[0]] = rulecolors[n[2]]; 
		var next = randomwalk(map,n[0],n[1],d-1);
		if(next != null){
			return next;
		} 
	}
	//If we don't have anything, back up
	return null;
}



//utility function 
function fill(length,filler){
	var a = new Array();
	for (var i = 0; i < length; i++){
		if(filler instanceof Array){a.push(filler.slice(0))}
		else{a.push(filler)}
	}
	return a;
}

//Fisher-Yates Shuffle
function fys(a){
	for(var i=a.length-1;i>0; i--){
		//6 is the number of colors
		var j = r() * i;
		j = ~~j; //Take the floor
		var tmp = a[j];
		a[j] = a[i];
		a[i] = tmp;
	}
	return a; //Meaningless since we edited the object, but just in case
}

function offset(square){
	if(square === undefined || square === null) return null;
	square = square.toLowerCase();
	if(square == "g") return 0;
	if(square == "m") return 1;
	if(square == "b") return 2;
	if(square == "w") return 3;
	if(square == "f") return 4;
	if(square == "d") return 5;
	if(square == "s") return 6;
}

//Game code
window.onload = function () {
	enchant();

	var game = new Game(res * 10, res * 8);
	game.mute = false;
	//Make it accessible for browser debugging
	this.theGame = game;
	game.fps = 30;
	game.scale = 1;
	//Graphics
	game.preload("map.big.png", "thog.big.png", "bunny.png");
	//Sounds
	game.preload("disco.mp3", "laugh.mp3");
	//Keybindings
	game.keybind(90,"a");
	game.keybind(88,"b");
	var paused = false;

	var Bunny = enchant.Class.create(enchant.Sprite, {initialize: function(x,y){
		//Add a bunny at the specified point. 
		//"Bunny" is a relative term.
		enchant.Sprite.call(this,res/4,res/4); //He's a little guy!
		this.x = x;
		this.y = y;
		this.image = game.assets["bunny.png"];

		this.addEventListener('enterframe', function(e){
			if(paused){return;}
			if(game.frame % 30 < 15){
				//TODO switch sprite
				this.frame = 1;
			}
			else {this.frame = 0;}
		})

	}});

	game.onload = function () {
		player = new Sprite(res,res);
		player.x = res * 5;
		player.y = res * 5;
		player.facing = 0;
		player.image = game.assets["thog.big.png"];

		player.addEventListener('enterframe', function(e){
			if(paused){return;}

			player.frame = player.direction * 3 + player.walk;
			//Handle borders
			if (player.isMoving) {
				player.moveBy(player.vx * 4 , player.vy * 4);
		 
				if (!(game.frame % 3)) {
					player.walk++;
					player.walk %= 3;
				}
				if ((player.vx && player.x % res == 0) || (player.vy && player.y % res == 0)) {
					player.isMoving = false;
					player.walk = 1;
				}
			} else {
				player.vx = player.vy = 0;
				if (game.input.left) {
					player.direction = 1;
					player.vx = -4;
					if(-1 === validrules[0].indexOf(rules[gmap[~~(player.y/res)][~~(player.x / res) -1]-2])){
						goback();
						return;
					}
				} else if (game.input.right) {
					player.direction = 2;
					player.vx = 4;
					if(-1 === validrules[2].indexOf(rules[gmap[ ~~(player.y/ res)][~~(player.x / res) +1]-2])){
						goback();
						return;
					}
				} else if (game.input.up) {
					player.direction = 3;
					player.vy = -4;
					if(-1 === validrules[1].indexOf(rules[gmap[ ~~(player.y/ res)-1][~~(player.x / res) ]-2])){
						goback();
						return;
					}
				} else if (game.input.down) {
					player.direction = 0;
					player.vy = 4;
					if(-1 === validrules[3].indexOf(rules[gmap[ ~~(player.y/ res)+1][~~(player.x / res) ]-2])){
						goback();
						return;
					}
				}
				if (player.vx || player.vy) {
					var x = player.x + (player.vx ? player.vx / Math.abs(player.vx) * res : 0);
					var y = player.y + (player.vy ? player.vy / Math.abs(player.vy) * res : 0);
					if (0 <= x && x < map.width && 0 <= y && y < map.height  && !map.hitTest(x, y)) {
							player.isMoving = true;
							arguments.callee.call(player);
					}
				}
			}

			if(game.frame % (30 * 5) === 0 && Math.random() < .33){
				//periodic sound effect
				//game.assets["shrimp.wav"].play();
			}
		});

		game.addEventListener(Event.A_BUTTON_DOWN, function(e){
			if(!game.pushStart){
				game.pushStart = true;
				game.clearEventListener('enterframe');
				startlevel(1);
			}
		});

		game.addEventListener(Event.B_BUTTON_DOWN, function(e){
		});

		document.addEventListener('keydown', function(e,f){
			if(e.keyCode == 77){ console.log("muting");game.mute = !(game.mute);}
			if(e.keyCode == 80){paused = !paused;}
		});

		var goback = function(){
			if(!game.mute){
				//Deal with sound cutting out...
				var tl = game.assets["laugh.mp3"].clone();
				tl.play();
				delete tl;
			}

			var bx = player.x + res/2 - res/8;
			var by = player.y + res/2 - res/8;
			//Use the sign to move the bunny!
			if(player.vx != 0){bx += (player.vx / Math.abs(player.vx)) * (res/4)};
			if(player.vy != 0){by += (player.vy / Math.abs(player.vy)) * (res/4)};
			var bunny = new Bunny(bx, by);
			game.currentScene.addChild(bunny);
			player.vx = 0;
			player.vy = 0;
			player.x = res * player.start.x;
			player.y = res * player.start.y;
		}

		var startlevel = function(level){
			//If this is the first level, generate rules
			var caption;
			if(level === 0){
				//attract mode
				game.rainbowLoop = 1;
				
					
				var rainbowSpin = function(){ 
					var fc = game.frame;
					if(fc % 15 != 0){return;}
					var map = new Map(res,res);
					map.image = game.assets["map.big.png"];
					var mapfiller = rainbow.slice(0);
					game.rainbowLoop++;
					game.rainbowLoop = game.rainbowLoop % 6;
					for(var i = 0;i<mapfiller.length;i++){
						for(var j = 0; j<mapfiller[0].length;j++){
							mapfiller[i][j] = ((mapfiller[i][j] + game.rainbowLoop ) % 6) + 2;
						}
					}
					map.loadData(mapfiller);
					var scene = new Scene();
					scene.addChild(map)
					game.pushScene(scene);
				}
				game.addEventListener('enterframe', rainbowSpin);
				return;
			} else if(level === 1){
				caption = "START THE BOOGIE";
				player.x = 5 * res;
				player.y = 5 * res;
				player.start = {x:5, y:5};
				for(var i=0;i<14;i++){rules.push(i+1);}
				fys(rules);
				rules = rules.slice(0,6);
				//Start the music!
			} else {
			//Otherwise, replace rule one at random, making sure 
			//each direction still has a valid rule
				caption = "GROOVE&nbsp;" + level;
				player.start = {x:player.x / res, y: player.y / res};
				var newrules = [];
				for(var i=0;i<14;i++){newrules.push(i);}
				newrules = _.filter(newrules, function(x){return (-1 === rules.indexOf(x))});
				fys(newrules);
				fys(validrules);
				for(var i=0;i<validrules.length;i++){
					if(validrules[i].length > 1){
						//OK, put in the new rule
						fys(validrules[i]);
						rules[rules.indexOf(validrules[0])] = newrules[0];
						break;
					}
				}
			}
			validrules = getvalidrules(rules);
			for(var i=1;i<15;i++){
				rulecolors[i] = rules.indexOf(i) + 2;
			}
				
			//Create the object...
			map = new Map(res,res);
			map.image = game.assets["map.big.png"];
			var mapdata = fill(world.h,fill(world.w,0));
			
			//Then get the next goal point and the walk
			mapdata[player.y / res][player.x / res] = 1;
			goalxy = randomwalk(mapdata, ~~(player.x / res), ~~(player.y / res), 25);

			//Then fill the rest of the map randomly
			for(i=0;i<mapdata.length;i++){
				for(j=0;j<mapdata[0].length;j++){
					if (mapdata[i][j] == 0){
						//There are six colors and they start at 2
						mapdata[i][j] = ~~(r() * 6) + 2;
					}
				}
			}


			map.loadData(mapdata);
			gmap = mapdata;
			map.x = 0;
			map.y = 0;
			//No collision for now
			var col = fill(world.h,fill(world.w,0));
			map.collisionData = col;
			var scene = new Scene();
			scene.addChild(map);
			var goal = new Goal(goalxy[0],goalxy[1]);
			scene.addChild(goal);
			scene.addChild(player);
			scene.backgroundColor = "#f0f";

			var stage = new Group();
			stage.addChild(map);
			stage.addChild(goal);
			stage.addChild(player);
			stage.x = 0;
			stage.y = 0;

			//And now add the level text
			var leveltext = new Scene();
			var label = new Label(caption);
			label.color = "#000"; //It would be awesome to make this rainbow, but...
			//label.width = game.width;
			label.y = game.height / 2;
			label.x = (game.width / 2) - (label.width / 2);
			label.font = "bold 50px sans-serif";
			leveltext.addChild(label);
			leveltext.countdown = 30;
			leveltext.addEventListener('enterframe',function(e){
                                //This should only display for a few seconds
				if(this.countdown===0) {paused = false; game.popScene();return;}
                                paused = true;
                                this.countdown--;
                        });


			scene.addEventListener("enterframe", function(e){
				//Camera follows player
				//stage.x = (game.width / 2) - player.x;
				//stage.y = (game.height / 2) - player.y;
				if(player.x == goal.x && player.y == goal.y){
					//Set next goal!
					startlevel(level+1);
				}
				
				if(game.mute){game.assets["disco.mp3"].volume = 0;}
				else {game.assets["disco.mp3"].volume = 1;}
				//Loop the bg music
				if(game.assets["disco.mp3"].currentTime == game.assets["disco.mp3"].duration || game.assets["disco.mp3"].currentTime == 0){
					game.assets["disco.mp3"].play();
				}

			});

			game.popScene();
			game.pushScene(scene);
			game.pushScene(leveltext);

		}

		var Goal = enchant.Class.create(enchant.Sprite, {initialize: function(x,y){
			enchant.Sprite.call(this,res,res);
			this.image = game.assets["map.big.png"];
			this.x = res * x;
			this.y = res * y;
			this.addEventListener('enterframe',function(e){
				if(game.frame % 30 < 15){
					this.opacity = 0;
				} else { this.opacity = 100; }
			});
		}});

		startlevel(0);
	};

	//logic 
	game.start();
}

